#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include "btserverdefines.h"
#include "btsocket.h"
#include "btutils.h"

#define SETTIMEOUT(timeout, tv, ptv) \
	if (timeout < 0) { \
		ptv = NULL; \
	} else { \
		if (timeout < 1000000) { \
			tv.tv_sec = 0; \
			tv.tv_usec = timeout; \
		} else { \
			tv.tv_sec = timeout / 1000000; \
			tv.tv_usec = timeout % 1000000; \
		} \
		ptv = &tv; \
	}

int BTSocketCompare(cBTSocket sk, const char* bdaddr, int chan, int *res)
{
	bdaddr_t addr;

	if (!sk || !bdaddr)
		return BT_ERR_ARGUMENT;
	str2ba(bdaddr, &addr);
	if (!bacmp(&addr, &sk->bdaddr) && sk->chan == chan)
		*res = BT_TRUE;
	else
		*res = BT_FALSE;

	return BT_ERR_NONE;
}

int BTSocketCreate(BTSocket* sk, const char* bdaddr, int chan)
{
	BTSocket s;

	if (!bdaddr || chan > 30)
		return BT_ERR_ARGUMENT;
	s = (BTSocket)malloc(sizeof(struct sBTSocket));
	if (!s)
		return BT_ERR_NOMEMORY;

	str2ba(bdaddr, &s->bdaddr);
	s->chan = chan;
	s->sock = -1;
	s->accsock = -1;
        s->tty_id = BTS_TTY_DEFAULT_ID;
	sprintf(s->tty, "%s", "");
	*sk = s;

	return BT_ERR_NONE;
}

int BTSocketDestroy(BTSocket sk)
{
	if (!sk)
		return BT_ERR_ARGUMENT;
	if (sk->sock > 0)
		close(sk->sock);
	if (sk->accsock > 0)
		close(sk->accsock);
	if (strlen(sk->tty))
		BTSocketReleaseTTY(sk);
	free(sk);

	return BT_ERR_NONE;
}

int BTSocketListen(BTSocket sk, int chan, int profile)
{
  int s = -1;
  bdaddr_t bdaddr;
  bacpy(&bdaddr, BDADDR_ANY);

  if (!sk)
    return BT_ERR_ARGUMENT;
  if (sk->sock > 0 || sk->accsock > 0)
    return BT_ERR_INUSE;

  if (profile == BT_PROFILE_AVRCP) // AVCT uses L2CAP
  {
    struct sockaddr_l2 addr;
    struct l2cap_options opts;
    int err;
    socklen_t optlen;
    
    memset(&addr, 0, sizeof(addr));
    if ((s = socket(PF_BLUETOOTH, SOCK_SEQPACKET, BTPROTO_L2CAP)) < 0)
    {
      DEBUG_LOG1("[BT SOCKET] : creating socket: %s", strerror(errno));
      return BT_ERR_FAILED;
    }
    if (fcntl(s, F_SETFD, FD_CLOEXEC) < 0)
    {
      DEBUG_LOG1("[BT SOCKET] : setting FD_CLOEXEC failed: %s", strerror(errno));
      close(s);
      return BT_ERR_FAILED;
    }
    addr.l2_family = AF_BLUETOOTH;
    bacpy(&addr.l2_bdaddr, &bdaddr);
    addr.l2_psm = htobs(L2CAP_PSM_AVCTP);
    if (bind(s, (struct sockaddr *)&addr, sizeof(addr)) < 0)
    {
      DEBUG_LOG1("[BT SOCKET] : binding socket: %s", strerror(errno));
      close(s);
      return BT_ERR_FAILED;
    }
    memset(&opts, 0, sizeof(opts));
    opts.imtu = L2CAP_DEFAULT_MTU;
    opts.omtu = L2CAP_DEFAULT_MTU;
    opts.flush_to = 0xffff;
    optlen = sizeof(opts);

    // MTU size negotiation  
    err = getsockopt(s, SOL_L2CAP, L2CAP_OPTIONS, &opts, &optlen );
    if( !err && ((opts.imtu > L2CAP_DEFAULT_MTU) || (opts.omtu > L2CAP_DEFAULT_MTU)) ) {
      opts.imtu = L2CAP_DEFAULT_MTU;
      opts.omtu = L2CAP_DEFAULT_MTU;
      opts.flush_to = 0xffff;
      setsockopt(s, SOL_L2CAP, L2CAP_OPTIONS, &opts, sizeof(opts));
    }
  }
  else // not avct, use RFCOMM
  {
    struct sockaddr_rc addr;

    memset(&addr, 0, sizeof(addr));
    if ((s = socket(PF_BLUETOOTH, SOCK_STREAM, BTPROTO_RFCOMM)) < 0)
    {
      DEBUG_LOG1("[BT SOCKET] : creating socket: %s", strerror(errno));
      return BT_ERR_FAILED;
    }
    if (fcntl(s, F_SETFD, FD_CLOEXEC) < 0)
    {
      DEBUG_LOG1("[BT SOCKET] : setting FD_CLOEXEC failed: %s", strerror(errno));
      close(s);
      return BT_ERR_FAILED;
    }
    addr.rc_family = AF_BLUETOOTH;
    hci_devba(0, &bdaddr);
    bacpy(&addr.rc_bdaddr, &bdaddr);
    addr.rc_channel = chan;

    if (bind(s, (struct sockaddr *)&addr, sizeof(addr)) < 0)
    {
      DEBUG_LOG1("[BT SOCKET] : binding socket: %s", strerror(errno));
      close(s);
      return BT_ERR_FAILED;
    }
  }
  // start listening
  if (listen(s, 1) < 0) {
    DEBUG_LOG1("[BT SOCKET] : listening socket: %s", strerror(errno));
    close(s);
    return BT_ERR_FAILED;
  }
  sk->accsock  = s;

  return BT_ERR_NONE;
}

int BTSocketStopListen(BTSocket sk)
{
	if (!sk)
		return BT_ERR_ARGUMENT;
	if (sk->accsock > 0)
		close(sk->accsock);
	sk->accsock = -1;

	return BT_ERR_NONE;
}

int BTSocketAccept(BTSocket sk, int timeout, int profile)
{
	fd_set rset;
	struct timeval *ptv;
	struct timeval tv, tvstart, tvstop, tvreq;
	socklen_t socklen;
	int res;

	if (!sk)
		return BT_ERR_ARGUMENT;
	if (sk->accsock < 0)
		return BT_ERR_ARGUMENT;
	if (sk->sock > 0)
		return BT_ERR_INUSE;

	SETTIMEOUT(timeout, tv, ptv);
	tvreq = tv;
retry:
	FD_ZERO(&rset);
	FD_SET(sk->accsock, &rset);
	if (ptv)
		gettimeofday(&tvstart, NULL);
	res = select(sk->accsock + 1, &rset, NULL, NULL, ptv);
	if (res < 0 || res > 1)
  {
		if (res < 0) {
			DEBUG_LOG1("[BT SOCKET] : select failed with %s", strerror(errno));
		} else {
			DEBUG_LOG1("[BT SOCKET] : select failed %d events pending", res);
		}
		return BT_ERR_FAILED;
	} else if (res == 0)
	{
		return BT_ERR_TIMEOUT;
	}
  if (profile == BT_PROFILE_AVRCP) // AVCT uses L2CAP
  {
    struct sockaddr_l2 addr;
    memset(&addr, 0, sizeof(addr));
    socklen = sizeof(addr);
    res = accept(sk->accsock, (struct sockaddr *) &addr, &socklen);
    if (res < 0)
    {
      DEBUG_LOG1("[BT SOCKET] : accept failed with %s", strerror(errno));
      return BT_ERR_FAILED;
    }
  }
  else // not avct, use RFCOMM
  {
    struct sockaddr_rc addr;
    memset(&addr, 0, sizeof(addr));
    socklen = sizeof(addr);
    res = accept(sk->accsock, (struct sockaddr *) &addr, &socklen);
    if (res < 0)
    {
      DEBUG_LOG1("[BT SOCKET] : accept failed with %s", strerror(errno));
      return BT_ERR_FAILED;
    }
    /* check if the connect is coming from the right bt device */
    if (!(bacmp(&sk->bdaddr, &addr.rc_bdaddr)==0) && !(bacmp(&sk->bdaddr, BDADDR_ANY)==0)) // not the same and not BDADDR_ANY (listening)
    {
      DEBUG_LOG("connected to a wrong device");
      close(res);
      /* calculate remained timeout */
      if (!ptv)
        goto retry;
      gettimeofday(&tvstop, NULL);
      tv.tv_sec = tvstop.tv_sec - tvstart.tv_sec;
      if (tvstop.tv_usec > tvstart.tv_usec)
        tv.tv_usec = tvstop.tv_usec - tvstart.tv_usec;
      else
        tv.tv_usec = tvstart.tv_usec - tvstop.tv_usec;
      if (tv.tv_sec < tvreq.tv_sec || (tv.tv_sec == tvreq.tv_sec && tv.tv_usec < tvreq.tv_usec)) {
        tv.tv_sec = tvreq.tv_sec - tv.tv_sec;
        tv.tv_usec = (tv.tv_usec < tvreq.tv_usec) ? tvreq.tv_usec - tv.tv_usec : tv.tv_usec - tvreq.tv_usec;
        tvreq = tv;
        goto retry;
      }
      return BT_ERR_TIMEOUT;
    }
    else
    {
      if (bacmp(&sk->bdaddr, BDADDR_ANY)==0)
      {
        DEBUG_LOG("BTSocketAccept : connected from listener -> copy address");
        bacpy(&sk->bdaddr, &addr.rc_bdaddr);
      }
    }
  }
	sk->sock = res;
	close(sk->accsock);
	sk->accsock = -1;

	return BT_ERR_NONE;
}

int BTSocketReceive(cBTSocket sk, char *buf, int len, int* received, int timeout)
{
	fd_set rset;
	struct timeval *ptv;
	struct timeval tv;
	int res, nr;

	if (!sk)
		return BT_ERR_ARGUMENT;
	if (sk->sock < 0 || !buf || len < 0 || !received)
		return BT_ERR_ARGUMENT;

	SETTIMEOUT(timeout, tv, ptv);
	FD_ZERO(&rset);
	FD_SET(sk->sock, &rset);
	res = select(sk->sock + 1, &rset, NULL, NULL, ptv);
	if (res < 0 || res > 1) {
		DEBUG_LOG1("[BT SOCKET] : select() failed: %s", strerror(errno));
		return BT_ERR_FAILED;
	}
	else if (res == 0)
		return BT_ERR_TIMEOUT;
	nr = recv(sk->sock, buf, len, MSG_NOSIGNAL);
	if (nr < 0) {
		DEBUG_LOG1("[BT SOCKET] : recv() failed: %s", strerror(errno));
		return BT_ERR_FAILED;
	}

	if (nr == 0 && !BTSocketIsConnected ( sk, BT_TO_NONE) ) {
		DEBUG_LOG("[BT_SOCKET] : recv() failed no connection!");
		return BT_ERR_FAILED;
	}

	*received = nr;

	return BT_ERR_NONE;
}

int BTSocketConnect(BTSocket sk, int timeout, int profile)
{
 	int s = -1;
  int res = -1;

 	if (!sk)
 		return BT_ERR_ARGUMENT;
 	if (sk->sock > 0 || sk->accsock > 0)
 		return BT_ERR_INUSE;

	if (profile == BT_PROFILE_AVRCP) // AVCT uses L2CAP
  {
  	struct sockaddr_l2 addr;

  	if ((s = socket(PF_BLUETOOTH, SOCK_SEQPACKET, BTPROTO_L2CAP)) < 0) {
  		DEBUG_LOG1("[BT_SOCKET] : error creating socket: %s", strerror(errno));
  		return BT_ERR_FAILED;
  	}
  	res = fcntl(s, F_GETFL, 0);
  	if (res < 0 || fcntl(s, F_SETFL, res | O_NONBLOCK) < 0) {
  		DEBUG_LOG1("[BT_SOCKET] : error setting non-blocking: %s", strerror(errno));
  		close(s);
  		return BT_ERR_FAILED;
  	}
  	if (fcntl(s, F_SETFD, FD_CLOEXEC) < 0) {
  		DEBUG_LOG1("[BT_SOCKET] : error setting FD_CLOEXEC failed: %s", strerror(errno));
  		close(s);
  		return BT_ERR_FAILED;
  	}
  	memset(&addr, 0, sizeof(addr));
  	addr.l2_family = AF_BLUETOOTH;
  	bacpy(&addr.l2_bdaddr, &sk->bdaddr);
  	addr.l2_psm = L2CAP_PSM_AVCTP;
  	if (!connect(s, (struct sockaddr *)&addr, sizeof(addr))) 
    {
  	  DEBUG_LOG("[BT_SOCKET] : BTSocketConnect L2CAP socket connected.");
    	sk->sock = s;
  		return BT_ERR_NONE;
  	}
  }
  else // not avct, use RFCOMM
  {
    struct sockaddr_rc addr;

  	if ((s = socket(PF_BLUETOOTH, SOCK_STREAM, BTPROTO_RFCOMM)) < 0) {
  		DEBUG_LOG1("[BT_SOCKET] : error creating socket: %s", strerror(errno));
  		return BT_ERR_FAILED;
  	}
  	res = fcntl(s, F_GETFL, 0);
  	if (res < 0 || fcntl(s, F_SETFL, res | O_NONBLOCK) < 0) {
  		DEBUG_LOG1("[BT_SOCKET] : error setting non-blocking: %s", strerror(errno));
  		close(s);
  		return BT_ERR_FAILED;
  	}
  	if (fcntl(s, F_SETFD, FD_CLOEXEC) < 0) {
  		DEBUG_LOG1("[BT_SOCKET] : error setting FD_CLOEXEC failed: %s", strerror(errno));
  		close(s);
  		return BT_ERR_FAILED;
  	}

  	memset(&addr, 0, sizeof(addr));
  	addr.rc_family = AF_BLUETOOTH;
  	bacpy(&addr.rc_bdaddr, &sk->bdaddr);
  	addr.rc_channel = sk->chan;
  	if (!connect(s, (struct sockaddr *)&addr, sizeof(addr))) 
    {
  		sk->sock = s;
		  DEBUG_LOG("[BT_SOCKET] : BTSocketConnect RFCOMM socket connected.");
      return BT_ERR_NONE;
  	}
	}
#ifdef NASHVILLE_BLUETOOTH
  if ((errno != EAGAIN) && (errno != EINPROGRESS))
#else
  // Do we need to distiguish non-nashville from nashville
  if (errno != EAGAIN)
#endif
  {
      DEBUG_LOG2("[BT_SOCKET] : connect to channel %d failed: %s",sk->chan, strerror(errno));
      close(s);
      return BT_ERR_FAILED;
  }
  sk->sock = s;

  res = BTSocketIsConnected(sk, timeout);
  if (res == BT_ERR_FAILED) {
  	close(sk->sock);
  	sk->sock = -1;
  	DEBUG_LOG("[BT_SOCKET] : BTSocketConnect socket connect FAILED.");
  }
  else
  {
    DEBUG_LOG("[BT_SOCKET] : BTSocketConnect socket CONNECTED.");
  }
  return res;
}


int BTSocketIsConnected(cBTSocket sk, int timeout)
{
	fd_set wset;
	struct timeval *ptv;
	struct timeval tv;
	int res, err;
	socklen_t optlen;

	if (!sk)
		return BT_ERR_ARGUMENT;
	if (sk->sock < 0)
		return BT_ERR_ARGUMENT;

	SETTIMEOUT(timeout, tv, ptv);
	FD_ZERO(&wset);
	FD_SET(sk->sock, &wset);
	res = select(sk->sock + 1, NULL, &wset, NULL, ptv);
	if (res < 0 || res > 1) {
		DEBUG_LOG1("[BT SOCKET] : select failed %s", strerror(errno));
		return BT_ERR_FAILED;
	}
	if (res == 0)
		return BT_ERR_TIMEOUT;
	optlen = sizeof(err);
	res = getsockopt(sk->sock, SOL_SOCKET, SO_ERROR, (void*)&err, &optlen);
	if (res < 0) {
		DEBUG_LOG1("[BT SOCKET] : getsockopt failed: %s", strerror(errno));
		return BT_ERR_FAILED;
	}
	if (err != 0) {
		DEBUG_LOG1("[BT SOCKET] : connect failed %s", strerror(err));
		return BT_ERR_FAILED;
	}
  
	return BT_ERR_NONE;
}

int BTSocketSend(cBTSocket sk, const char* buf, int len, int *sent, int timeout)
{
	fd_set wset;
	struct timeval *ptv;
	struct timeval tv;
	int res;
	ssize_t ns;

	if (!sk)
		return BT_ERR_ARGUMENT;
	if (sk->sock < 0 || !buf || len < 0 || !sent)
		return BT_ERR_ARGUMENT;

	FD_ZERO(&wset);
	FD_SET(sk->sock, &wset);

	SETTIMEOUT(timeout, tv, ptv);

	res = select(sk->sock + 1, NULL, &wset, NULL, ptv);
	if (res < 0 || res > 1) {
		DEBUG_LOG1("[BT SOCKET] : select failed %s", strerror(errno));
		return BT_ERR_FAILED;
	}
	else if (res == 0)
		return BT_ERR_TIMEOUT;

	ns = send(sk->sock, buf, len, MSG_NOSIGNAL);
	if (ns < 0) {
		DEBUG_LOG1("[BT SOCKET] : send failed %s", strerror(errno));
		return BT_ERR_FAILED;
	}
	if (ns == 0 && !BTSocketIsConnected ( sk, BT_TO_NONE) ) {
		DEBUG_LOG("[BT SOCKET] : send() failed no connection!\n");
		return BT_ERR_FAILED;
	}

	*sent = ns;
	return BT_ERR_NONE;
}

int BTSocketClose(BTSocket sk)
{
	if (!sk)
		return BT_ERR_ARGUMENT;

	close(sk->sock);
	sk->sock = -1;
	close(sk->accsock);
	sk->accsock = -1;

	return BT_ERR_NONE;
}

int BTSocketCreateTTY(BTSocket sk, int tty_id)
{
	int ctl;
	struct rfcomm_dev_req req;
	char bdstr[BT_BDADDR_LEN];
        char ttyname[BTS_TTY_NAME_LEN];

	if (!sk)
		return BT_ERR_ARGUMENT;

        /* valid ids are [0-3] */  
        if( (tty_id < 0) || (tty_id > 3))
          return BT_ERR_ARGUMENT;

        sprintf(ttyname, BTS_TTY_TEMPL, tty_id);

	ba2str(&sk->bdaddr, bdstr);
        DEBUG_LOG3("[BT SOCKET] : creating %s for %s, channel %d", ttyname, bdstr, sk->chan);

	if ((ctl = socket(AF_BLUETOOTH, SOCK_RAW, BTPROTO_RFCOMM)) < 0) {
		DEBUG_LOG1("[BT SOCKET] : creating control socket: %s", strerror(errno));
		return BT_ERR_FAILED;
	}
	memset(&req, 0, sizeof(req));
	req.dev_id = tty_id;
	bacpy(&req.src, BDADDR_ANY);
	bacpy(&req.dst, &sk->bdaddr);
	req.channel = sk->chan;
	if (ioctl(ctl, RFCOMMCREATEDEV, &req) < 0) {
		DEBUG_LOG1("[BT SOCKET] : ioctl: %s", strerror(errno));
		close(ctl);
		return BT_ERR_FAILED;
	}
	close(ctl);
	sprintf(sk->tty, ttyname);
        sk->tty_id = tty_id;

	return BT_ERR_NONE;
}

int BTSocketIsTTYReady(cBTSocket sk, int *ready)
{
	struct stat st;

	if (!sk)
		return BT_ERR_ARGUMENT;
	if (stat(sk->tty, &st) < 0) {
			*ready = BT_FALSE;
	} else
		*ready = BT_TRUE;

	return BT_ERR_NONE;
}

int BTSocketReleaseTTY(BTSocket sk)
{
	int ctl;
	struct rfcomm_dev_req req;

	if (!sk)
		return BT_ERR_ARGUMENT;
	memset(&req, 0, sizeof(req));
	req.dev_id = sk->tty_id;
	if ((ctl = socket(AF_BLUETOOTH, SOCK_RAW, BTPROTO_RFCOMM)) < 0) {
		DEBUG_LOG1("[BT SOCKET] : creating control socket: %s", strerror(errno));
		return BT_ERR_FAILED;
	}
	if (ioctl(ctl, RFCOMMRELEASEDEV, &req)  < 0) {
		DEBUG_LOG1("[BT SOCKET] : ioctl: %s", strerror(errno));
		close( ctl );
		return BT_ERR_FAILED;
	}
	close(ctl);
	sprintf(sk->tty, "%s", "");

	return BT_ERR_NONE;
}

int BTSocketGetTTYName(cBTSocket sk, char* ttyname)
{
	if (!sk)
		return BT_ERR_ARGUMENT;
	strcpy(ttyname, sk->tty);

	return BT_ERR_NONE;
}

int BTSocketGetAddress(cBTSocket sk, char *addr)
{
	if (!sk)
		return BT_ERR_ARGUMENT;
	ba2str(&sk->bdaddr, addr);
  
	return BT_ERR_NONE;
}

int BTSocketGetChannel(cBTSocket sk, int *chan)
{
	if (!sk)
		return BT_ERR_ARGUMENT;
	*chan = sk->chan;

	return BT_ERR_NONE;
}

int BTSocketGetMtuSize(cBTSocket sk, int * imtu, int * omtu)
{
  if (!imtu || !omtu)
    return BT_ERR_ARGUMENT;

  struct l2cap_options opts;
  memset(&opts, 0, sizeof(opts));
  socklen_t opts_size = sizeof(opts);
  int rc = getsockopt(sk->sock,
                      SOL_L2CAP,
                      L2CAP_OPTIONS,
                      &opts,
                      &opts_size);

  if (rc==-1)
  {
    char rc_str[12];
    switch(rc)
    {
    case EBADF:        strcpy(rc_str, "EBADF");        break;
    case EINVAL:       strcpy(rc_str, "EINVAL");       break;
    case ENOPROTOOPT:  strcpy(rc_str, "ENOPROTOOPT");  break;
    case ENOTSOCK:     strcpy(rc_str, "ENOTSOCK");     break;
    case EACCES:       strcpy(rc_str, "EACCES");       break;
    case ENOBUFS:      strcpy(rc_str, "ENOBUFS");      break;
    default:           strcpy(rc_str, "OK");           break;
  }

    DEBUG_LOG2("Unable to query socket options. err=%d (%s)", errno, rc_str);
    return BT_ERR_FAILED;
  }

  *imtu = opts.imtu;
  *omtu = opts.omtu;

  DEBUG_LOG4("l2cap options: imtu=%d, omtu=%d, mode=%d, flush_to=%d.", opts.imtu, opts.omtu, opts.mode, opts.flush_to);

  return BT_ERR_NONE;
}
