/*
 *
 *  BlueZ - Bluetooth protocol stack for Linux
 *
 *  Copyright (C) 2003-2004  Marcel Holtmann <marcel@holtmann.org>
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation;
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 *  OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT OF THIRD PARTY RIGHTS.
 *  IN NO EVENT SHALL THE COPYRIGHT HOLDER(S) AND AUTHOR(S) BE LIABLE FOR ANY
 *  CLAIM, OR ANY SPECIAL INDIRECT OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES 
 *  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN 
 *  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF 
 *  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *  ALL LIABILITY, INCLUDING LIABILITY FOR INFRINGEMENT OF ANY PATENTS, 
 *  COPYRIGHTS, TRADEMARKS OR OTHER RIGHTS, RELATING TO USE OF THIS 
 *  SOFTWARE IS DISCLAIMED.
 *
 *
 *  $Id: main.c,v 1.15 2004/12/25 17:43:17 holtmann Exp $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <malloc.h>
#include <syslog.h>
#include <signal.h>
#include <getopt.h>
#include <sys/poll.h>
#include <sys/ioctl.h>
#include <sys/socket.h>

#include <bluetooth/bluetooth.h>
#include <bluetooth/hci.h>
#include <bluetooth/hci_lib.h>
#include <bluetooth/l2cap.h>
#include <bluetooth/sdp.h>
#include <bluetooth/hidp.h>

#include "hidd.h"
#include "TTMP.h"
#include "btttmp.h"
#include "bthidremote.h"
#include "btutils.h"

#undef TTMP_ERROR
#define TTMP_ERROR

#define POLL_NO_TIMEOUT      0
#define TIMEOUT_CONNECTION   70 // minutes, timeout in bluetooth remote is 60 minutes

int hidState = HID_NO_CONNECTION;
char hidStarted = 0;
char hidRemoteAddress[HID_REMOTE_ADDRESS_SIZE];
char hidDeviceName[HID_REMOTE_NAME_SIZE];

static int ctl = -1, csk = -1, isk = -1, ncsk = -1, nisk = -1;
static bdaddr_t bdaddr; // bt address of connected device - used to disconnect later on

static int l2cap_listen(const bdaddr_t *bdaddr, unsigned short psm, int lm, int backlog)
{
	struct sockaddr_l2 addr;
	struct l2cap_options opts;
	int sk;

	if ((sk = socket(PF_BLUETOOTH, SOCK_SEQPACKET, BTPROTO_L2CAP)) < 0)
		return -1;

	memset(&addr, 0, sizeof(addr));
	addr.l2_family = AF_BLUETOOTH;
	bacpy(&addr.l2_bdaddr, bdaddr);
	addr.l2_psm = htobs(psm);

	if (bind(sk, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
		close(sk);
		return -1;
	}

	setsockopt(sk, SOL_L2CAP, L2CAP_LM, &lm, sizeof(lm));

	memset(&opts, 0, sizeof(opts));
	opts.imtu = HIDP_DEFAULT_MTU;
	opts.omtu = HIDP_DEFAULT_MTU;
	opts.flush_to = 0xffff;

	setsockopt(sk, SOL_L2CAP, L2CAP_OPTIONS, &opts, sizeof(opts));

	if (listen(sk, backlog) < 0) {
		close(sk);
		return -1;
	}
  fcntl(sk, F_SETFD,FD_CLOEXEC);

	return sk;
}

static int l2cap_accept(int sk, bdaddr_t *bdaddr)
{
	struct sockaddr_l2 addr;
	socklen_t addrlen;
	int nsk;

	memset(&addr, 0, sizeof(addr));
	addrlen = sizeof(addr);

	if ((nsk = accept(sk, (struct sockaddr *) &addr, &addrlen)) < 0)
		return -1;

	if (bdaddr)
		bacpy(bdaddr, &addr.l2_bdaddr);

	return nsk;
}

static int request_authentication(bdaddr_t *src, bdaddr_t *dst)
{
	struct hci_conn_info_req *cr;
	char addr[18];
	int err, dd, dev_id;

	ba2str(src, addr);
	dev_id = hci_devid(addr);
	if (dev_id < 0)
		return dev_id;

	dd = hci_open_dev(dev_id);
	if (dd < 0)
		return dd;

	cr = malloc(sizeof(*cr) + sizeof(struct hci_conn_info));
	if (!cr)
		return -ENOMEM;

	bacpy(&cr->bdaddr, dst);
	cr->type = ACL_LINK;
	err = ioctl(dd, HCIGETCONNINFO, (unsigned long) cr);
	if (err < 0) {
		free(cr);
		hci_close_dev(dd);
		return err;
	}

	err = hci_authenticate_link(dd, htobs(cr->conn_info->handle), 25000);

	free(cr);
	hci_close_dev(dd);

	return err;
}

static int request_encryption(bdaddr_t *src, bdaddr_t *dst)
{
	struct hci_conn_info_req *cr;
	char addr[18];
	int err, dd, dev_id;

	ba2str(src, addr);
	dev_id = hci_devid(addr);
	if (dev_id < 0)
		return dev_id;

	dd = hci_open_dev(dev_id);
	if (dd < 0)
		return dd;

	cr = malloc(sizeof(*cr) + sizeof(struct hci_conn_info));
	if (!cr)
		return -ENOMEM;

	bacpy(&cr->bdaddr, dst);
	cr->type = ACL_LINK;
	err = ioctl(dd, HCIGETCONNINFO, (unsigned long) cr);
	if (err < 0) {
		free(cr);
		hci_close_dev(dd);
		return err;
	}

	err = hci_encrypt_link(dd, htobs(cr->conn_info->handle), 1, 25000);

	free(cr);
	hci_close_dev(dd);

	return err;
}

static int create_device(int ctl, int csk, int isk, uint8_t subclass, int nosdp, int encrypt, int timeout)
{
	struct hidp_connadd_req req;
	struct sockaddr_l2 addr;
	socklen_t addrlen;
	bdaddr_t src, dst;
	char bda[18];
	int err;

	memset(&addr, 0, sizeof(addr));
	addrlen = sizeof(addr);

	if (getsockname(csk, (struct sockaddr *) &addr, &addrlen) < 0)
		return -1;

	bacpy(&src, &addr.l2_bdaddr);

	memset(&addr, 0, sizeof(addr));
	addrlen = sizeof(addr);

	if (getpeername(csk, (struct sockaddr *) &addr, &addrlen) < 0)
		return -1;

	bacpy(&dst, &addr.l2_bdaddr);

	memset(&req, 0, sizeof(req));
	req.ctrl_sock = csk;
	req.intr_sock = isk;
	req.flags     = 0;
	req.idle_to   = timeout * 60;

	if (!nosdp) {
		err = get_hid_device_info(&src, &dst, &req);
		if (err < 0)
			goto error;
	} else {
		struct l2cap_conninfo conn;
		socklen_t size;
		uint8_t class[3];

		memset(&conn, 0, sizeof(conn));
		size = sizeof(conn);
		if (getsockopt(csk, SOL_L2CAP, L2CAP_CONNINFO, &conn, &size) < 0)
			memset(class, 0, 3);
		else
			memcpy(class, conn.dev_class, 3);

		if (class[1] == 0x25 && (class[2] == 0x00 || class[2] == 0x01))
			req.subclass = class[0];
		else
			req.subclass = 0xc0;
	}

	if (subclass != 0x00)
		req.subclass = subclass;

	ba2str(&dst, bda);
	DEBUG_LOG2("New HID device %s (%s)", bda, req.name);

	if (encrypt && (req.subclass & 0x40)) {
		err = request_authentication(&src, &dst);
		if (err < 0) {
			DEBUG_LOG1("Authentication for %s failed", bda);
			goto error;
		}

		err = request_encryption(&src, &dst);
		if (err < 0)
			DEBUG_LOG1("Encryption for %s failed", bda);
	}

	err = ioctl(ctl, HIDPCONNADD, &req);

error:
	if (req.rd_data)
		free(req.rd_data);

	return err;
}

//
// from hcitool -name 
// 
static void cmd_name(bdaddr_t bdaddr, char *name, int len)
{
	int dd;
	int dev_id = 0;

	if (dev_id < 0) {
		dev_id = hci_get_route(&bdaddr);
		if (dev_id < 0) {
			DEBUG_LOG("Device is not available.\n");
			return;
		}
	}

	dd = hci_open_dev(dev_id);
	if (dd < 0) {
		DEBUG_LOG("Device is not available.\n");
		return;
	}

	if (hci_read_remote_name(dd, &bdaddr, len, name, 1000) == 0)
	  DEBUG_LOG1("********* cmd_name %s\n", name);

	close(dd);	
	return;
}

void stop_listening()
{
    // close listen sockets
    DEBUG_LOG("close isk and csk sockets to stop listening");
    if( isk >= 0)
    {
      DEBUG_LOG("close isk");
      close(isk);
      isk = -1;
    }
    if( csk >= 0)
    {
      DEBUG_LOG("close cks");
      close(csk);
      csk = -1;
    }
}

void close_nsockets_device()
{
    // close listen sockets
    DEBUG_LOG("close nisk and incsk sockets to stop listening");
    if( nisk >= 0)
    {
      DEBUG_LOG("close nisk");
      close(nisk);
      nisk = -1;
    }
    if( ncsk >= 0)
    {
      DEBUG_LOG("close ncsk");
      close(ncsk);
      ncsk = -1;
    }

}

int run_server()
{
  struct pollfd p[2];
  char addr[18];
  short events;
  int err;

  p[0].fd = csk;
  p[0].events = POLLIN | POLLERR | POLLHUP;

  p[1].fd = isk;
  p[1].events = POLLIN | POLLERR | POLLHUP;

  p[0].revents = 0;
  p[1].revents = 0;

  err = poll(p, sizeof(p)/sizeof(p[0]), POLL_NO_TIMEOUT); //non blocking
  if (err == 0)
  {
  	return 0;
    //	continue due to timeout (ie. no data available)
  }
  else
  {	
    if (err < 0)
	{
	  return err;
	}

    // Somebody wants to connect, get the name
    events = p[0].revents | p[1].revents;    // set input events.
  
    if (events & POLLIN) 
    {
      ncsk = l2cap_accept(csk, &bdaddr);
      nisk = l2cap_accept(isk, NULL);
      
      if(hidState == HID_CONNECTED)
      {
        DEBUG_LOG1("Disconnect: %s because already connected\n", hidRemoteAddress);
        close_nsockets_device();
      }
      else
      {
        DEBUG_LOG("Stop listening\n");
        stop_listening();
    
        cmd_name(bdaddr, hidDeviceName, 248);
        ba2str(&bdaddr, addr);
    
        strncpy( hidRemoteAddress, addr, 18);
        hidState = HID_CONNECTION_PENDING;
        DEBUG_LOG3("HID_CONNECTION_PENDING hidState:%d hidDeviceName:%s hidRemoteAddress:%s\n", hidState, hidDeviceName, hidRemoteAddress);
      }
    }
	else
	{
		DEBUG_LOG1("HIDD Unhandled event %d\n", events);
    }
  }
  return 0;
}

int start_listening()
{
  int err = BT_ERR_NONE;
  int lm = 0;
  bacpy(&bdaddr, BDADDR_ANY);

  if(csk >=0) {
  	close(csk);
	csk = -1;
  }

  if(isk >=0) {
  	close(isk);
	isk = -1;
  }
  
  csk = l2cap_listen(&bdaddr, L2CAP_PSM_HIDP_CTRL, lm, 1);
  if (csk < 0) {
    perror("Can't listen on HID control channel");
    if( ctl >= 0)
    {
      close(ctl);
      ctl = -1;
    }
    hidState = HID_FAILED;
    err = -1;
  }

  isk = l2cap_listen(&bdaddr, L2CAP_PSM_HIDP_INTR, lm, 1);
  if (isk < 0) {
    perror("Can't listen on HID interrupt channel");
    if( ctl >= 0)
    {
      close(ctl);
      ctl = -1;
    }
    if( csk >= 0)
    {
      close(csk);
      csk = -1;
    }
    hidState = HID_FAILED;
    err = -1;
  }

  return err;
}

int connect_device()
{
  int result = BT_ERR_FAILED;
  int err = create_device(ctl, ncsk, nisk, 0x40, 1, 0, TIMEOUT_CONNECTION);
  
  DEBUG_LOG3("Create_device after, ctl:%d ncsk:%d nisk:%d", ctl, ncsk, nisk);
  if (err < 0)
  {
	DEBUG_LOG2("HID create error %d (%s)", errno, strerror(errno));
  }
  else
  {
    hidState = HID_CONNECTED;
    result = BT_ERR_NONE;
  }

  close_nsockets_device();

  return result;
}

void close_ctl()
{
  if( ctl >= 0)
  {
    close(ctl);
    ctl = -1;
  }
}

int start_ctl()
{
  int err = BT_ERR_NONE;
  ctl = socket(AF_BLUETOOTH, SOCK_RAW, BTPROTO_HIDP);
  if (ctl < 0) {
    perror("Can't open HIDP control socket"); 
    hidState = HID_FAILED;
    err = -1;
  }

  return err;
}

int disconnect_device()
{ 
  struct hidp_conndel_req req;
  int err = BT_ERR_NONE;

  DEBUG_LOG("start disconnect device!!!\n");
  if(bacmp(&bdaddr, BDADDR_ANY) != 0)
  {
    // ie have bt address then disconnect
	bacpy(&req.bdaddr, &bdaddr);
	req.flags = (1 << HIDP_VIRTUAL_CABLE_UNPLUG);
	err = ioctl(ctl, HIDPCONNDEL, &req);
	DEBUG_LOG1("Warning ioclt disconnect failed %d\n", err);
  }

  stop_listening();
  close_nsockets_device();

#ifdef BT_HID_REMOTE 
  BTHidRemoteHandleInit();
#endif // BT_HID_REMOTE

  close_ctl();
 
  hidStarted = 0;
  err = BT_ERR_NONE;
  DEBUG_LOG("end disconnect device!!!\n");
  
  return err;
}

int start_hid()
{
  int err = BT_ERR_NONE;
  if( !hidStarted) 
  {
    DEBUG_LOG("@@@@@@@@@@@@@@@@@@@@@ start_hid \n");
    if( start_ctl() == BT_ERR_NONE)
    {
      if( start_listening() == BT_ERR_NONE)
      {
        hidStarted = 1;
        hidState = HID_NO_CONNECTION;
      }
      else
      { 
        err = BT_ERR_FAILED;
      }
    }
    else
    {
      err = BT_ERR_FAILED; 
    }
  }
  else
  { 
    DEBUG_LOG1("@@@@@@@@@@@@@@@@@@@@@ start_hid already started! hidState:%d\n", hidState);
  }
  
  return err;
}

static void usage(void)
{
	printf("hidd - Bluetooth HID daemon version %s\n\n", "2.15");     // hardcode version!

	printf("Usage:\n"
		"\thidd [options] [commands]\n"
		"\n");

	printf("Options:\n"
		"\t-i <hciX|bdaddr>     Local HCI device or BD Address\n"
		"\t-t <timeout>         Set idle timeout (in minutes)\n"
		"\t-b <subclass>        Overwrite the boot mode subclass\n"
		"\t-n, --nodaemon       Don't fork daemon to background\n"
		"\t-h, --help           Display help\n"
		"\n");

	printf("Commands:\n"
		"\t--server             Start HID server\n"
		"\t--search             Search for HID devices\n"
		"\t--connect <bdaddr>   Connect remote HID device\n"
		"\t--unplug <bdaddr>    Unplug the HID connection\n"
		"\t--kill <bdaddr>      Terminate HID connection\n"
		"\t--killall            Terminate all connections\n"
		"\t--show               List current HID connections\n"
		"\n");
}

static struct option main_options[] = {
	{ "help",	0, 0, 'h' },
	{ "nodaemon",	0, 0, 'n' },
	{ "subclass",	1, 0, 'b' },
	{ "timeout",	1, 0, 't' },
	{ "device",	1, 0, 'i' },
	{ "master",	0, 0, 'M' },
	{ "encrypt",	0, 0, 'E' },
	{ "nosdp",	0, 0, 'D' },
	{ "show",	0, 0, 'l' },
	{ "list",	0, 0, 'l' },
	{ "server",	0, 0, 'd' },
	{ "listen",	0, 0, 'd' },
	{ "search",	0, 0, 's' },
	{ "create",	1, 0, 'c' },
	{ "connect",	1, 0, 'c' },
	{ "disconnect",	1, 0, 'k' },
	{ "terminate",	1, 0, 'k' },
	{ "release",	1, 0, 'k' },
	{ "kill",	1, 0, 'k' },
	{ "killall",	0, 0, 'K' },
	{ "unplug",	1, 0, 'u' },
	{ 0, 0, 0, 0 }
};

int hidd(int argc, char *argv[])
{
	bdaddr_t bdaddr, dev;
	uint32_t flags = 0;
	uint8_t subclass = 0x00;
	int log_option = LOG_NDELAY | LOG_PID | LOG_PERROR;
	int opt;
	int mode = 0, daemon = 1, nosdp = 0, encrypt = 0, timeout = 30, lm = 0;

	bacpy(&bdaddr, BDADDR_ANY);
	// openlog("hidd", log_option, LOG_DAEMON);

	while ((opt = getopt_long(argc, argv, "+i:nt:b:MEDldsc:k:Ku:h", main_options, NULL)) != -1) {
		switch(opt) {
		case 'i':
			if (!strncasecmp(optarg, "hci", 3))
				hci_devba(atoi(optarg + 3), &bdaddr);
			else
				str2ba(optarg, &bdaddr);
			break;
		case 'n':
			daemon = 0;
			break;
		case 't':
			timeout = atoi(optarg);
			break;
		case 'b':
			if (!strncasecmp(optarg, "0x", 2))
				subclass = (uint8_t) strtol(optarg, NULL, 16);
			else
				subclass = atoi(optarg);
			break;
		case 'M':
			lm |= L2CAP_LM_MASTER;
			break;
		case 'E':
			encrypt = 1;
			break;
		case 'D':
			nosdp = 1;
			break;
		case 'l':
			mode = 0;
			break;
		case 'd':
			mode = 1;
			break;
		case 's':
			mode = 2;
			break;
		case 'c':
			str2ba(optarg, &dev);
			mode = 3;
			break;
		case 'k':
			str2ba(optarg, &dev);
			mode = 4;
			break;
		case 'K':
			bacpy(&dev, BDADDR_ALL);
			mode = 4;
			break;
		case 'u':
			str2ba(optarg, &dev);
			flags = (1 << HIDP_VIRTUAL_CABLE_UNPLUG);
			mode = 4;
			break;
		case 'h':
			usage();
      break;
		default:
      DEBUG_LOG1("Unknown opt:%c\n", opt);
      break;
		}
	}

  start_hid();

  log_option |= LOG_PERROR;

	DEBUG_LOG("before run_server \n");

	// printf("run_server ctl:%d\ncsk:%d\nisk:%d\nsubclass:0x%x\nnodsp:%d\nencrypt:%d\ntimeout%d\n", ctl, csk, isk, subclass, nosdp, encrypt, timeout);

	DEBUG_LOG("Exit");

	return 0;
}
